import gwyfile
import gsffile
import numpy as np
import pandas as pd
import os
from pathlib import PurePath
import re


class Reader:
    def __init__(self, fullfilepath=None):
        self.filename = fullfilepath


class GwyReader(Reader):
    def __init__(self, fullfilepath=None, channelname=None):
        super().__init__(fullfilepath)
        self.channelname = channelname

    def read(self):
        # Returns a dictionary of all the channels
        gwyobj = gwyfile.load(self.filename)
        allchannels = gwyfile.util.get_datafields(gwyobj)

        if self.channelname is None:
            return allchannels
        else:
            # Read channels from gwyfile and return only a specific one
            channel = allchannels[self.channelname]
            return channel


class GsfReader(Reader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        data, metadata = gsffile.read_gsf(self.filename)
        channel = gwyfile.objects.GwyDataField(
            data,
            xreal=metadata["XReal"],
            yreal=metadata["YReal"],
            xoff=metadata["XOffset"],
            yoff=metadata["YOffset"],
            si_unit_xy=None,
            si_unit_z=None,
            typecodes=None,
        )
        return channel


class NeaHeaderReader(Reader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    @staticmethod
    def parseline(linestring, params={}):
        ct = linestring.split("\t")
        fieldname = ct[0][2:-1]
        fieldname = fieldname.replace(" ", "")

        if "Scanner Center Position" in linestring:
            fieldname = fieldname[:-5]
            params[fieldname] = [float(ct[2]), float(ct[3])]

        elif "Scan Area" in linestring:
            fieldname = fieldname[:-7]
            params[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

        elif "Pixel Area" in linestring:
            fieldname = fieldname[:-7]
            params[fieldname] = [int(ct[2]), int(ct[3]), int(ct[4])]

        elif "Averaging" in linestring:
            params[fieldname] = int(ct[2])

        elif "Interferometer Center/Distance" in linestring:
            fieldname = fieldname.replace("/", "")
            params[fieldname] = [
                float(ct[2].replace(",", "")),
                float(ct[3].replace(",", "")),
            ]

        elif "Regulator" in linestring:
            fieldname = fieldname[:-7]
            params[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

        elif "Q-Factor" in linestring:
            fieldname = fieldname.replace("-", "")
            params[fieldname] = float(ct[2])

        else:
            fieldname = ct[0][2:-1]
            fieldname = fieldname.replace(" ", "")
            val = ct[2]
            val = val.replace(",", "")
            try:
                params[fieldname] = float(val)
            except:
                params[fieldname] = val.strip()

        return params

    def read(self):
        params = {}
        with open(self.filename, encoding="utf8") as f:
            # Read www.neaspec.com
            line = f.readline()
            count = 1
            while f:
                line = f.readline()
                count = count + 1
                try:
                    if line[0] not in ("#", "\n"):
                        break
                    if line[0] == "#":
                        params = NeaHeaderReader.parseline(line, params)
                except IndexError:
                    break

            channels = line.strip().split("\t")
            channels = [channel.strip() for channel in channels]

        return channels, params


class NeaInfoReader(NeaHeaderReader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        _, infodict = super().read()
        return infodict


class NeaSpectralReader(Reader):
    def __init__(self, fullfilepath=None, output="dict"):
        super().__init__(fullfilepath)
        self._output = output

    def read(self):
        data = {}

        channels, params = NeaHeaderReader(self.filename).read()
        channels.append("")

        count = len(list(params.keys())) + 2

        data = pd.read_csv(
            self.filename,
            sep="\t",
            skiprows=count,
            encoding="utf-8",
            names=channels,
            lineterminator="\n",
        ).dropna(axis=1, how="all")

        cols_to_keep = [c for c in data.columns if c != ""]
        data = data[cols_to_keep]

        if self._output == "dict":
            data = data.to_dict("list")
            for key in list(data.keys()):
                data[key] = np.asarray(data[key])

        return data, params


class ImageStackReader(Reader):
    """Reads a list of images from the subfolders of the specified folder by loading the files that contain the pattern string in the filename"""

    def __init__(self, folder=None, folder_pattern=""):
        super().__init__(folder)
        self.folder = self.filename
        self.folder_pattern = folder_pattern

    def read(self, pattern):
        imagestack = []
        wns = []
        filepaths = get_filenames(
            self.folder, pattern, folderpattern=self.folder_pattern
        )

        for i, path in enumerate(filepaths):
            data_reader = GsfReader(path)
            imagestack.append(data_reader.read().data)

            try:
                txtpath = recreate_infofile_name_from_path(path)
                inforeader = NeaInfoReader(txtpath)
                infodict = inforeader.read()
                wn = get_wl_from_infofile(infodict)
                wns.append(wn)
            except:
                wns.append(i)

        idxs = np.argsort(np.asarray(wns))
        imagestack = [imagestack[i] for i in idxs]
        wns = [wns[i] for i in idxs]

        return imagestack, wns


def get_wl_from_infofile(infodict: dict):
    """Returns the wavelength from the info file. If not found, returns 0.0"""
    try:
        if infodict["TargetWavelength"] == "":
            wn = infodict["InterferometerCenterDistance"][0]
        else:
            wn = infodict["TargetWavelength"]
            # NeaSpec is not consistent in the units of the wavelength.
            # Sometimes it is in cm-1, sometimes in um.
            if wn < 50.0:
                wn = 10000 / wn
    except:
        wn = None
    return wn


def get_wl_from_filename(filename):
    """Returns the wavelength from the filename. If not found, returns None"""

    wn = re.findall(
        r"(?<=[_-])(\d+(?:\.\d+)?)(?=(?:_?cm[-_]1|-?cm[-_]1))", PurePath(filename).name
    )

    if wn:
        wn = float(wn[0])
    else:
        wn = None

    return wn


def get_filenames(folder: str, pattern: str, folderpattern=""):
    """Returns the filepath of all files in the subfolders of the specified folder that contain pattern string in the filename"""

    filepaths = []

    for subfolder in os.listdir(folder):
        if os.path.isdir(os.path.join(folder, subfolder)) and re.search(
            folderpattern, subfolder
        ):
            for name in os.listdir(os.path.join(folder, subfolder)):
                if re.search(pattern, name):
                    subpath = os.path.join(subfolder, name)
                    filepaths.append(os.path.join(folder, subpath))

    return filepaths


def recreate_infofile_name_from_path(filepath: str):
    """Recreates the name of the info file from the path of the data file"""

    pathparts = list(PurePath(filepath).parts)
    newparts = pathparts[:-1]
    newparts.append(pathparts[-2] + ".txt")

    return str(PurePath(*newparts))


class NeaFileLegacyReader(Reader):
    """Reader for .nea files from older neasnom microscopes"""

    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        data = {}
        params = {}

        with open(self.filename, encoding="utf8") as f:
            h = next(f)  # header
            h = h.strip()
            h = h.split("\t")

            l = next(f)
            l = l.strip()
            l = l.split("\t")

            f.seek(0)
            next(f)
            datacols = np.arange(len(h), len(l))
            C_data = np.loadtxt(f, dtype="float", usecols=datacols)

            f.seek(0)
            next(f)

            metacols = np.arange(0, 4)
            meta = np.loadtxt(
                f,
                dtype={"names": tuple(h), "formats": (int, int, int, "S10")},
                usecols=metacols,
            )
            if "Run" in h:
                runs = np.unique(meta["Run"])
            else:
                runs = [0]

            Max_row = len(np.unique(meta["Row"]))
            Max_col = len(np.unique(meta["Column"]))
            Max_run = len(runs)
            Max_omega = np.shape(C_data)[1]

            N_rows = Max_row * Max_col * Max_run * Max_omega

            indexes = np.unique(meta["Channel"], return_index=True)[1]
            channels = [meta["Channel"][index] for index in sorted(indexes)]
            channels = [channel.decode("utf-8") for channel in channels]

            for name in h:
                if name != "Channel":
                    data[name] = np.array(meta[name])

            for i in range(len(channels)):
                data[channels[i]] = np.ravel(C_data[i * Max_run : (i + 1) * Max_run, :])

        alpha = 0
        beta = 0
        data["Run"] = np.zeros(N_rows)
        data["Column"] = np.zeros(N_rows)
        data["Row"] = np.zeros(N_rows)

        for i in range(0, N_rows, Max_omega * Max_run):
            if beta == Max_row:
                beta = 0
                alpha = alpha + 1
            data["Run"][i : i + Max_omega * Max_run] = np.repeat(
                np.arange(Max_run), Max_omega
            )
            data["Column"][i : i + Max_omega * Max_run] = alpha
            data["Row"][i : i + Max_omega * Max_run] = beta
            beta = beta + 1

            params["PixelArea"] = [
                Max_row,
                Max_col,
                Max_omega,
            ]
            params["Scan"] = "Fourier Scan"

        return data, params



import unittest
import os

import numpy as np

import pySNOM
from pySNOM import readers, spectra


class test_Neaspectrum(unittest.TestCase):
    def test_pointspectrum_object(self):
        f = "datasets/testspectrum_singlepoint.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        s = spectra.NeaSpectrum(data, params)

        np.testing.assert_almost_equal(s.data["O2A"][0], 0.1600194)
        np.testing.assert_string_equal(s.parameters["Scan"], "Fourier Scan")
        np.testing.assert_string_equal(s.scantype, "Point")
        np.testing.assert_equal(np.shape(s.data["O2A"])[0], 2048)

    def test_multipointspectrum_object(self):
        f = "datasets/testspectrum_multipoint.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        s = spectra.NeaSpectrum(data, params)

        np.testing.assert_almost_equal(s.data["O2A"][1, 0, 0], 0.1600194)
        np.testing.assert_string_equal(s.parameters["Scan"], "Fourier Scan")
        np.testing.assert_string_equal(s.scantype, "LineScan")
        np.testing.assert_equal(np.shape(s.data["O2A"])[2], 4)

    def test_transfromations(self):
        f = "datasets/testspectrum_singlepoint.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()
        fref = "datasets/testspectrum_singlepoint_ref.txt"
        file_reader_ref = readers.NeaSpectralReader(
            os.path.join(pySNOM.__path__[0], fref)
        )
        data_ref, params_ref = file_reader_ref.read()

        s = spectra.NeaSpectrum(data, params)
        r = spectra.NeaSpectrum(data_ref, params_ref)

        channel = "O2A"
        normdata = spectra.NormalizeSpectrum(spectra.DataTypes.Amplitude).transform(
            s.data[channel], r.data[channel]
        )
        corrdata = spectra.LinearNormalize(
            wavenumber1=1000, wavenumber2=2200, datatype=spectra.DataTypes.Amplitude
        ).transform(normdata, s.data["Wavenumber"])

        np.testing.assert_almost_equal(normdata[1000], 0.7278023)
        np.testing.assert_almost_equal(corrdata[1000], 0.9795999)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from pySNOM.images import (
    LineLevel,
    BackgroundPolyFit,
    MaskedBackgroundPolyFit,
    SimpleNormalize,
    DataTypes,
    AlignImageStack,
    RemoveSpikes,
    ValueFillIn,
    ScarRemoval,
    mask_from_datacondition,
    dict_from_imagestack,
)


class TestLineLevel(unittest.TestCase):
    def test_median(self):
        d = np.arange(12).reshape(3, -1)[:, [0, 1, 3]]
        l = LineLevel(method="median", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(
            out, [[-1.0, 0.0, 2.0], [-1.0, 0.0, 2.0], [-1.0, 0.0, 2.0]]
        )
        l = LineLevel(method="median", datatype=DataTypes.Amplitude)
        out = l.transform(d)
        np.testing.assert_almost_equal(
            out, [[0.0, 1.0, 3.0], [0.8, 1.0, 1.4], [0.8888889, 1.0, 1.2222222]]
        )

    def test_mean(self):
        d = np.arange(12).reshape(3, -1)[:, [0, 1, 3]]
        l = LineLevel(method="mean", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(
            out,
            [
                [-1.3333333, -0.3333333, 1.6666667],
                [-1.3333333, -0.3333333, 1.6666667],
                [-1.3333333, -0.3333333, 1.6666667],
            ],
        )
        l = LineLevel(method="mean", datatype=DataTypes.Amplitude)
        out = l.transform(d)
        np.testing.assert_almost_equal(
            out,
            [
                [0.0, 0.75, 2.25],
                [0.75, 0.9375, 1.3125],
                [0.8571429, 0.9642857, 1.1785714],
            ],
        )

    def test_difference(self):
        d = np.arange(12).reshape(3, -1)[:, [0, 1, 3]]
        l = LineLevel(method="difference", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(
            out, [[0.0, 1.0, 3.0], [0.0, 1.0, 3.0], [0.0, 1.0, 3.0]]
        )
        l = LineLevel(method="difference", datatype=DataTypes.Amplitude)
        out = l.transform(d)
        np.testing.assert_almost_equal(
            out, [[0.0, 1.0, 3.0], [0.8, 1.0, 1.4], [0.88888889, 1.0, 1.22222222]]
        )

    def test_masking_mean(self):
        d = np.zeros([8, 10])
        d[2:6, 3:7] = 1
        mask = np.ones([8, 10])
        mask[2:6, 3:7] = np.nan

        l = LineLevel(method="mean", datatype=DataTypes.Phase)

        out = l.transform(d)
        np.testing.assert_almost_equal(out[5, 0], -0.4)
        out = l.transform(d, mask=mask)
        np.testing.assert_almost_equal(out[5, 0], 0.0)

    def test_masking_median(self):
        d = np.zeros([8, 10])
        d[2:6, 2:9] = 1
        mask = np.ones([8, 10])
        mask[2:6, 2:9] = np.nan

        l = LineLevel(method="median", datatype=DataTypes.Phase)

        out = l.transform(d)
        np.testing.assert_almost_equal(out[5, 0], -1.0)
        out = l.transform(d, mask=mask)
        np.testing.assert_almost_equal(out[5, 0], 0.0)

    def test_difference_masking(self):
        d = np.zeros([8, 10])
        d[2:6, 2:9] = 1
        mask = np.ones([8, 10])
        mask[2:6, 2:9] = np.nan

        l = LineLevel(method="difference", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(out[5, 0], -1.0)
        np.testing.assert_almost_equal(out[1, 0], 0.0)
        np.testing.assert_almost_equal(out[5, 2], 0.0)


class TestBackgroundPolyFit(unittest.TestCase):
    def test_withmask(self):
        d = np.ones([10, 10])
        d[4:8, 4:8] = 10
        mask = np.ones([10, 10])
        mask[4:8, 4:8] = np.nan

        t = MaskedBackgroundPolyFit(xorder=1, yorder=1, datatype=DataTypes.Phase)
        out = t.transform(d, mask=mask)
        np.testing.assert_almost_equal(out[0, 0], 0.0)
        np.testing.assert_almost_equal(out[9, 9], 0.0)

        t = MaskedBackgroundPolyFit(xorder=1, yorder=1, datatype=DataTypes.Amplitude)
        out = t.transform(d, mask=mask)
        np.testing.assert_almost_equal(out[0, 0], 1.0)
        np.testing.assert_almost_equal(out[9, 9], 1.0)

    def test_withoutmask(self):
        d = np.ones([10, 10])
        d[4:8, 4:8] = 10
        mask = np.ones([10, 10])
        mask[4:8, 4:8] = np.nan

        t = MaskedBackgroundPolyFit(xorder=1, yorder=1, datatype=DataTypes.Phase)
        out = t.transform(d)
        np.testing.assert_almost_equal(out[0, 0], -0.2975206611570238)
        np.testing.assert_almost_equal(out[9, 9], -3.439338842975202)

        t = MaskedBackgroundPolyFit(xorder=1, yorder=1, datatype=DataTypes.Amplitude)
        out = t.transform(d)
        np.testing.assert_almost_equal(out[0, 0], 0.7707006369426758)
        np.testing.assert_almost_equal(out[9, 9], 0.22525876833718098)

    def test_old_version(self):
        d = np.ones([10, 10])
        d[4:8, 4:8] = 10
        mask = np.ones([10, 10])
        mask[4:8, 4:8] = np.nan

        t = BackgroundPolyFit(xorder=1, yorder=1, datatype=DataTypes.Phase)
        out, _ = t.transform(d)
        np.testing.assert_almost_equal(out[0, 0], -0.2975206611570238)
        np.testing.assert_almost_equal(out[9, 9], -3.439338842975202)

        t = BackgroundPolyFit(xorder=1, yorder=1, datatype=DataTypes.Amplitude)
        out, _ = t.transform(d)
        np.testing.assert_almost_equal(out[0, 0], 0.7707006369426758)
        np.testing.assert_almost_equal(out[9, 9], 0.22525876833718098)


class TestScarRemoval(unittest.TestCase):
    def test_scarremove(self):
        data = np.tile(np.linspace(0, 0.5, 6), (5, 1)).T
        data[3, 1:-1] = 0.9
        t = ScarRemoval(threshold=0.5)
        out = t.transform(data)

        np.testing.assert_almost_equal(out[3, 2], 0.30000000000000004)
        np.testing.assert_almost_equal(data[3, 2], 0.9)


class TestHelperFunction(unittest.TestCase):
    def test_mask_from_condition(self):
        d = np.ones([2, 2])
        d[0, 0] = 0
        out = mask_from_datacondition(d < 1)

        np.testing.assert_equal(out[0, 0], np.nan)


class TestSimpleNormalize(unittest.TestCase):
    def test_median(self):
        d = np.zeros([8, 10])
        d[2:9, 1:9] = 1
        mask = np.ones([8, 10])
        mask[2:9, 1:9] = np.nan

        l = SimpleNormalize(method="median", datatype=DataTypes.Phase)

        out = l.transform(d)
        np.testing.assert_almost_equal(out[0, 0], -1.0)
        out = l.transform(d, mask=mask)
        np.testing.assert_almost_equal(out[0, 0], 0.0)

    def test_mean(self):
        d = np.zeros([8, 10])
        d[2:9, 1:9] = 1
        mask = np.ones([8, 10])
        mask[2:9, 1:9] = np.nan

        l = SimpleNormalize(method="mean", datatype=DataTypes.Phase)

        out = l.transform(d)
        np.testing.assert_almost_equal(out[0, 0], -0.6)
        out = l.transform(d, mask=mask)
        np.testing.assert_almost_equal(out[0, 0], 0.0)

    def test_min(self):
        d = np.asarray([1.0, 2.0, 3.0])
        mask = np.asarray([np.nan, 1, 1])

        l = SimpleNormalize(method="min", datatype=DataTypes.Phase)

        out = l.transform(d)
        np.testing.assert_almost_equal(out, [0.0, 1.0, 2.0])
        out = l.transform(d, mask=mask)
        np.testing.assert_almost_equal(out, [-1.0, 0.0, 1.0])

class TestFillIn(unittest.TestCase):
    def test_value_fillin(self):
        d = np.ones([9, 9])
        mask = np.random.rand(9,9)
        mask[4,4] = np.nan

        l = ValueFillIn(mask=mask,value=np.inf)

        out = l.transform(d)
        np.testing.assert_equal(out[4,4],np.inf)
        
class TestRemoveSpikes(unittest.TestCase):
    def test_remove_laplace(self):
        d = np.ones([9, 9])
        d[4, 4] = 0.8

        l = RemoveSpikes(threshold=0.9,method='laplace')

        out = l.transform(d)
        np.testing.assert_almost_equal(out[4,4],1.0)

    def test_remove_higher_laplace(self):
        d = np.ones([9, 9])
        d[4, 4] = 1.2

        l = RemoveSpikes(threshold=1.1,method='laplace',higher=True)

        out = l.transform(d)
        np.testing.assert_almost_equal(out[4,4],1.0)

    def test_remove_manual(self):
        d = np.ones([9, 9])
        d[4, 4] = 0.8

        l = RemoveSpikes(threshold=0.9,method='manual',value=1.11111)

        out = l.transform(d)
        np.testing.assert_almost_equal(out[4,4],1.11111)

    def test_remove_median(self):
        d = np.ones([9, 9])
        d[4, 4] = 0.8

        l = RemoveSpikes(threshold=0.9,method='median')

        out = l.transform(d)
        np.testing.assert_almost_equal(out[4,4],1.0)

    def test_remove_nan(self):
        d = np.ones([9, 9])
        d[4, 4] = 0.8

        l = RemoveSpikes(threshold=0.9,method='asdfsdfhdfg')

        out = l.transform(d)
        np.testing.assert_almost_equal(out[4,4],np.nan)


class TestAlignImageStack(unittest.TestCase):
    def test_stackalignment(self):
        image1 = np.zeros((50, 100))
        image2 = np.zeros((50, 100))
        image1[10:40, 10:40] = 1
        image2[20:50, 20:50] = 1

        aligner = AlignImageStack()
        shifts, crossrect = aligner.calculate([image1, image2])
        np.testing.assert_equal(shifts, [np.asarray([-10.0, -10.0])])
        np.testing.assert_equal(crossrect, [10, 0, 40, 90])

        out = aligner.transform([image1, image2], shifts, crossrect)
        np.testing.assert_equal(np.shape(out), (2, 29, 90))


class TestHelperFunctions(unittest.TestCase):
    def test_dictfromimagestack(self):
        stack = [np.zeros((50, 100)), np.zeros((50, 100))]

        out, outparams = dict_from_imagestack(stack, "O2A")

        self.assertEqual(outparams["PixelArea"], [50, 100, 2])
        self.assertTrue("M" in list(out.keys()))


if __name__ == "__main__":
    unittest.main()



import numpy as np
from enum import Enum
from pySNOM.images import type_from_channelname
from pySNOM.defaults import Defaults

MeasurementModes = Enum(
    "MeasurementModes", ["None", "nanoFTIR", "PsHet", "PTE", "nanoRaman"]
)
DataTypes = Enum("DataTypes", ["Amplitude", "Phase", "Complex", "Topography"])
ChannelTypes = Enum("ChannelTypes", ["None", "Optical", "Mechanical"])
ScanTypes = Enum("ScanTypes", ["Point", "LineScan", "HyperScan"])


class NeaSpectrum:
    """Storing full spectral measurement data and corresponding measurement parameters"""

    def __init__(
        self,
        data: dict,
        parameters: dict,
        scantype="Point",
        filename=None,
        mode="nanoFTIR",
    ):
        self.filename = filename  # Full path with name
        self._parameters = parameters
        self.data = data

        # set measurement mode
        try:
            self._mode = MeasurementModes[mode]
        except ValueError:
            self._mode = MeasurementModes["nanoFTIR"]
            raise ValueError(mode + "is not a measurement mode!")

        try:
            self._scantype = ScanTypes[scantype]
        except ValueError:
            self._scantype = ScanTypes["Point"]
            raise ValueError(scantype + "is not a measurement mode!")

        if parameters:
            try:
                if parameters["PixelArea"][1] == 1 and parameters["PixelArea"][0] == 1:
                    self._scantype = ScanTypes["Point"]
                elif parameters["PixelArea"][1] == 1 or parameters["PixelArea"][0] == 1:
                    self._scantype = ScanTypes["LineScan"]
                else:
                    self._scantype = ScanTypes["HyperScan"]
                self._mode = MeasurementModes[
                    Defaults().spectral_mode_defs[parameters["Scan"]]
                ]
            except:
                raise ValueError("Parameters dictionary is not valid!")
        else:
            # set measurement mode
            try:
                self._mode = MeasurementModes[mode]
            except ValueError:
                self._mode = MeasurementModes["nanoFTIR"]
                raise ValueError(mode + "is not a measurement mode!")
            # set scan type
            try:
                self._scantype = ScanTypes[scantype]
            except ValueError:
                self._scantype = ScanTypes["Point"]
                raise ValueError(scantype + "is not a measurement mode!")

    @property
    def data(self):
        """Property - data (dict with measurement channels)"""
        return self._data

    @data.setter
    def data(self, value):
        """Data setter to reshape properly"""
        self._data = Tools.reshape_spectrum_data(value, self._parameters)

    @property
    def mode(self):
        """Property - mode (MeasurementMode Enum name)"""
        return self._mode.name

    @property
    def scantype(self):
        """Property - scantype (ScanType Enum name)"""
        return self._scantype.name

    @property
    def parameters(self):
        """Property - scantype (ScanType Enum name)"""
        return self._parameters


class SingleChannelSpectrum(NeaSpectrum):
    def __init__(
        self,
        data,
        wndata,
        filename=None,
        parameters=None,
        mode="nanoFTIR",
        channelname="O2A",
    ):
        super().__init__(data, filename, parameters, mode)
        self._wndata = wndata
        self.channel = channelname
        self.order = 0
        self.datatype = None

    @property
    def data(self):
        """Property - data (dict with measurement channels)"""
        return self._data

    @property
    def wndata(self):
        """Property - wavenumber data"""
        return self._wndata

    @property
    def channel(self):
        """Property - channel (string)"""
        return self._channel

    @channel.setter
    def channel(self, value):
        self._channel = value
        self.channeltype, self.order, self.datatype = type_from_channelname(value)


# TRANSFORMATIONS ------------------------------------------------------------------------------------------------------------------
class Transformation:
    def transform(self, data):
        raise NotImplementedError()


class LinearNormalize(Transformation):
    def __init__(self, wavenumber1=0.0, wavenumber2=1000.0, datatype=DataTypes.Phase):
        self.wn1 = wavenumber1
        self.wn2 = wavenumber2
        self.datatype = datatype

    def transform(self, spectrum, wnaxis):
        wn1idx = np.argmin(abs(wnaxis - self.wn1))
        wn2idx = np.argmin(abs(wnaxis - self.wn2))
        m = (spectrum[wn2idx] - spectrum[wn1idx]) / (wnaxis[wn2idx] - wnaxis[wn1idx])
        C = spectrum[wn1idx] - m * wnaxis[wn1idx]

        if self.datatype == DataTypes.Amplitude:
            return spectrum / (m * wnaxis + C)
        else:
            return spectrum - (m * wnaxis + C)


class RotatePhase(Transformation):
    def __init__(self, degree=0.0, wn_ref=1000.0):
        self.wn_ref = wn_ref
        self.degree = degree

    def transform(self, spectrum, wnaxis):
        if not np.iscomplex(spectrum).any():
            spectrum = np.exp(spectrum * complex(1j))

        angles = wnaxis * np.deg2rad(self.degree) / self.wn_ref

        return np.angle(spectrum * np.exp(angles * complex(1j)))


class ShiftPhaseToZero(Transformation):
    """
    Calculates and applies the phase rotation needed to get a flat, levelled phase spectrum.
    Two reference frequencies have to be provided.
    """

    def __init__(self, wavenumber1=1000.0, wavenumber2=2200.0):
        self.wn1 = wavenumber1
        self.wn2 = wavenumber2

    def transform(self, spectrum, wnaxis):
        if not np.iscomplex(spectrum).any():
            spectrum = np.exp(spectrum * complex(1j))

        wn1idx = np.argmin(abs(wnaxis - self.wn1))
        wn2idx = np.argmin(abs(wnaxis - self.wn2))

        theta1 = np.angle(spectrum[wn1idx])
        wn1 = wnaxis[wn1idx]
        spectrum = spectrum * np.exp(-theta1 * complex(1j))

        theta2 = np.angle(spectrum[wn2idx])
        wn2 = wnaxis[wn2idx]

        angles = (wnaxis - wn1) * theta2 / (wn2 - wn1)
        spectrum = np.angle(spectrum * np.exp(-1 * angles * complex(1j)))

        return spectrum


class NormalizeSpectrum(Transformation):
    def __init__(self, datatype=DataTypes.Phase, dounwrap=False):
        self.datatype = datatype
        self.dounwrap = dounwrap

    def transform(self, spectrum, refspectrum):
        if self.datatype == DataTypes.Phase or self.datatype == DataTypes.Topography:
            if self.dounwrap:
                return np.unwrap(spectrum - refspectrum)
            else:
                return spectrum - refspectrum
        else:
            return spectrum / refspectrum


# TOOLS ------------------------------------------------------------------------------------------------------------------
class Tools:
    @staticmethod
    def reshape_spectrum_data(data, params):
        # To compensate for the zero-filling that NeaSpec does
        n = 1
        if params["Scan"] == "Fourier Scan":
            n = 2

        allchannels = list(data.keys())
        if "Depth" in allchannels:
            spectral_depth = len(np.unique(data["Depth"]))
        elif "Index" in allchannels:
            spectral_depth = len(np.unique(data["Index"]))
        elif "Omega" in allchannels:
            spectral_depth = len(np.unique(data["Omega"]))
        else:
            spectral_depth = params["PixelArea"][2] * n

        for channel in allchannels:
            # Point spectrum
            if params["PixelArea"][1] == 1 and params["PixelArea"][0] == 1:
                data[channel] = np.reshape(data[channel], (spectral_depth))

            # Linescan and HyperScan
            else:
                data[channel] = np.reshape(
                    data[channel],
                    (
                        params["PixelArea"][0],
                        params["PixelArea"][1],
                        spectral_depth,
                    ),
                )

        return data



import unittest
import os

import numpy as np

import pySNOM
from pySNOM import readers
from pySNOM.interferograms import (
    NeaInterferogram,
    ProcessSingleChannel,
    ProcessMultiChannels,
    ProcessAllPoints,
)


class test_Neaspectrum(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        fdata = "datasets/testifg_singlepoint.txt"
        data_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], fdata))
        data, measparams = data_reader.read()
        self.ifg = NeaInterferogram(data, measparams, filename=fdata)

        f = "datasets/testifg_multipoints.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()
        self.multi_ifg = NeaInterferogram(data, params)

    def test_pointinterferogram_object(self):
        np.testing.assert_almost_equal(self.ifg.data["O2A"][100], 9.564073)
        np.testing.assert_string_equal(self.ifg.parameters["Scan"], "Fourier Scan")
        np.testing.assert_string_equal(self.ifg.scantype, "Point")
        np.testing.assert_equal(np.shape(self.ifg.data["O2A"])[0], 2048)

    def test_multipointinterferogram_object(self):
        np.testing.assert_array_equal(
            self.multi_ifg.data["Run"][1, 0], np.asarray([0, 0, 0, 1, 1, 1])
        )

    def test_singlechannel_process(self):
        a2, p2, wn2 = ProcessSingleChannel(order=2, simpleoutput=True).transform(
            self.ifg
        )

        np.testing.assert_almost_equal(a2[500], 62.18682449597315)
        np.testing.assert_almost_equal(p2[500], 2.7921248371008596)
        np.testing.assert_almost_equal(wn2[500], 1276.2183207594132)

    def test_multichannel_process(self):
        s_multichannel = ProcessMultiChannels(apod=True).transform(self.ifg)

        np.testing.assert_almost_equal(
            s_multichannel.data["O3A"][500], 23.374168641358338
        )
        np.testing.assert_almost_equal(
            s_multichannel.data["O3P"][500], -0.9999766071179981
        )
        np.testing.assert_almost_equal(
            s_multichannel.data["Wavenumber"][500], 1276.2183207594132
        )

    def test_fullauto_process(self):
        s_allpoints = ProcessAllPoints().transform(self.ifg)

        np.testing.assert_almost_equal(s_allpoints.data["O3A"][500], 23.374168641358338)
        np.testing.assert_almost_equal(
            s_allpoints.data["O3P"][500], -0.9999766071179981
        )
        np.testing.assert_almost_equal(
            s_allpoints.data["Wavenumber"][500], 1276.2183207594132
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import copy
from enum import Enum
from scipy import signal
from scipy.fft import fft, fftshift
from scipy.interpolate import CubicSpline, interp1d
from pySNOM.spectra import NeaSpectrum, SingleChannelSpectrum
import re

MeasurementModes = Enum("MeasurementModes", ["None", "nanoFTIR"])
DataTypes = Enum("DataTypes", ["Amplitude", "Phase", "Topography"])
ChannelTypes = Enum("ChannelTypes", ["None", "Optical", "Mechanical"])
ScanTypes = Enum("ScanTypes", ["Point", "LineScan", "HyperScan"])


# INTERFEROGRAMS ------------------------------------------------------------------------------------------------------------------
class NeaInterferogram(NeaSpectrum):
    def __init__(
        self, data, parameters, scantype="Point", filename=None, mode="nanoFTIR"
    ):
        super().__init__(
            data, parameters, scantype=scantype, filename=filename, mode=mode
        )
        self.data = data

    @property
    def data(self):
        """Property - data (dict with measurement channels)"""
        return self._data

    @data.setter
    def data(self, value):
        """Data setter to reshape properly"""
        self._data = Tools.reshape_ifg_data(value, self._parameters)

    def add_channel(self, values, channelname):
        """Adds a new channel to data dictionary"""
        if channelname not in list(self._data.keys()):
            self._data[channelname] = np.reshape(
                values,
                (
                    int(self.parameters["PixelArea"][0]),
                    int(self.parameters["PixelArea"][1]),
                    int(self.parameters["PixelArea"][2] * self.parameters["Averaging"]),
                ),
            )
        else:
            raise ValueError


# TRANSFORMATIONS ------------------------------------------------------------------------------------------------------------------
class Transformation:
    def transform(self, data):
        raise NotImplementedError()


class ProcessInterferogram(Transformation):
    def __init__(self, apod=True, windowtype="blackmanharris", nzeros=4, wlpidx=None):
        self.apod = apod
        self.nzeros = nzeros
        self.wlpidx = wlpidx
        self.windowtype = windowtype

    def transform(self, ifg, maxis):
        # Find the location index of the WLP
        if self.wlpidx is None:
            self.wlpidx = np.argmax(np.abs(ifg))
        # Create apodization window
        if self.apod:
            w = Tools.asymmetric_window(
                npoints=len(ifg), centerindex=self.wlpidx, windowtype=self.windowtype
            )
        else:
            w = np.ones(np.shape(ifg))
        ifg = ifg - np.mean(ifg)
        # Calculate FFT
        complex_spectrum = fftshift(fft(ifg * w, self.nzeros * len(ifg)))
        # Calculate frequency axis
        stepsizes = np.median(np.diff(maxis * 1e6))
        Fs = 1 / np.mean(stepsizes)
        faxis = (Fs / 2) * np.linspace(-1, 1, len(complex_spectrum)) * 10000 / 2
        return (
            complex_spectrum[int(len(faxis) / 2) :],
            faxis[int(len(faxis) / 2) :],
        )


class InterpolateInterferogram(Transformation):
    """Reinterpolates the raw interferograms to a uniform grid."""

    def __init__(self, method="spline"):
        self.method = method

    def transform(self, ifg, maxis):
        # if np.iscomplex(ifg).any():
        newifg = np.zeros(np.shape(ifg)) * complex(1j)
        # else:
        # newifg = np.zeros(np.shape(ifg))

        newmaxis = np.zeros(np.shape(maxis))

        match self.method:
            case "spline":
                interp_object = CubicSpline
            case "linear":
                interp_object = interp1d

        # in case of processing multiple interferograms in a 2d array we have to take the median
        # sometimes the point locations has large jumps at the beginning (neaspec machine artifact)
        startM = np.min(np.nanmedian(maxis, axis=0))
        stopM = np.max(np.nanmedian(maxis, axis=0))

        if ifg.ndim == 1:
            newcoords = np.linspace(startM, stopM, num=len(maxis))
            if np.iscomplex(ifg).any():
                interpR = interp_object(maxis, np.real(ifg))
                interpI = interp_object(maxis, np.imag(ifg))
                newifgR = interpR(newcoords)
                newifgI = interpI(newcoords)
                newifg = newifgR + newifgI * complex(1j)
            else:
                interpifg = interp_object(maxis, ifg)
                newifg = interpifg(newcoords)
            return newifg, newcoords
        elif ifg.ndim == 2:
            newcoords = np.linspace(startM, stopM, num=np.shape(maxis)[1])
            for i in range(np.shape(ifg)[0]):
                if np.iscomplex(ifg[i][:]).any():
                    interpR = interp_object(maxis[i][:], np.real(ifg[i][:]))
                    interpI = interp_object(maxis[i][:], np.imag(ifg[i][:]))
                    newifgR = interpR(newcoords)
                    newifgI = interpI(newcoords)
                    newifg[i][:] = newifgR + newifgI * complex(1j)
                    newmaxis[i][:] = newcoords
                else:
                    interpifg = interp_object(maxis[i][:], ifg[i][:])
                    newifg[i][:] = interpifg(newcoords)
                    newmaxis[i][:] = newcoords
            return newifg, newmaxis
        else:
            return ifg, maxis


class ProcessSingleChannel(Transformation):
    def __init__(
        self,
        order,
        method="complex",
        apod=True,
        windowtype="blackmanharris",
        nzeros=4,
        interpmethod="spline",
        simpleoutput=False,
    ):
        self.order = order
        self.method = method
        self.windowtype = windowtype
        self.apod = apod
        self.nzeros = nzeros
        self.interpmethod = interpmethod
        self.simpleoutput = simpleoutput

    def transform(self, neaifg):  # Load amplitude and phase of the given channel
        # Calculate the interferogram to process based on the given method

        channelA = f"O{self.order}A"
        channelP = f"O{self.order}P"

        ifgA = np.reshape(
            neaifg.data[channelA],
            (neaifg.parameters["Averaging"], neaifg.parameters["PixelArea"][2]),
        )
        ifgP = np.reshape(
            neaifg.data[channelP],
            (neaifg.parameters["Averaging"], neaifg.parameters["PixelArea"][2]),
        )
        Maxis = np.reshape(
            neaifg.data["M"],
            (neaifg.parameters["Averaging"], neaifg.parameters["PixelArea"][2]),
        )

        match self.method:
            case "abs":
                IFG = np.abs(ifgA * np.exp(ifgP * complex(1j)))
            case "real":
                IFG = np.real(ifgA * np.exp(ifgP * complex(1j)))
            case "imag":
                IFG = np.imag(ifgA * np.exp(ifgP * complex(1j)))
            case "complex":
                IFG = ifgA * np.exp(ifgP * complex(1j))
            case "simple":
                IFG = ifgA

        #  Interpolate
        IFG, Maxis = InterpolateInterferogram(method=self.interpmethod).transform(
            IFG, Maxis
        )

        # PROCESS IFGs
        # Check if it is multiple interferograms or just a single one
        if len(np.shape(IFG)) == 1:
            complex_spectrum, f = ProcessInterferogram(
                apod=self.apod, windowtype=self.windowtype, nzeros=self.nzeros
            ).transform(IFG, Maxis)
            amp = np.abs(complex_spectrum)
            phi = np.angle(complex_spectrum)
        else:
            # Allocate variables
            spectraAll = complex(1j) * np.zeros(
                (np.shape(IFG)[0], int(self.nzeros * np.shape(IFG)[1] / 2))
            )
            fAll = np.zeros(np.shape(spectraAll))
            # Go trough all
            for i in range(np.shape(IFG)[0]):
                spectraAll[i, :], fAll[i, :] = ProcessInterferogram(
                    apod=self.apod, windowtype=self.windowtype, nzeros=self.nzeros
                ).transform(IFG[i, :], Maxis[i, :])
            # Average the complex spectra
            complex_spectrum = np.mean(spectraAll, axis=0)
            # Extract amplitude and phase from the averaged complex spectrum
            amp = np.abs(complex_spectrum)
            phi = np.angle(complex_spectrum)
            f = np.mean(fAll, axis=0)

        if self.simpleoutput:
            return amp, phi, f
        else:
            spectrum_data = {}
            spectrum_parameters = copy.deepcopy(neaifg.parameters)
            spectrum_parameters["ScanArea"] = [
                neaifg.parameters["ScanArea"][0],
                neaifg.parameters["ScanArea"][1],
                len(amp),
            ]
            spectrum_data[channelA] = amp
            spectrum_data[channelP] = phi
            spectrum_data["Wavenumber"] = f
            spectrum = NeaSpectrum(spectrum_data, spectrum_parameters)

            return spectrum


class ProcessMultiChannels(Transformation):
    def __init__(
        self,
        method="complex",
        apod=True,
        windowtype="blackmanharris",
        nzeros=4,
        interpmethod="spline",
        simpleoutput=False,
    ):
        self.method = method
        self.apod = apod
        self.windowtype = windowtype
        self.nzeros = nzeros
        self.interpmethod = interpmethod
        self.simpleoutput = simpleoutput

    def transform(self, neaifg):
        spectrum_data = {}
        spectrum_parameters = copy.deepcopy(neaifg.parameters)

        for order in range(6):
            channelA = f"O{order}A"
            channelP = f"O{order}P"

            amp, phi, f = ProcessSingleChannel(
                order,
                method=self.method,
                apod=self.apod,
                windowtype=self.windowtype,
                nzeros=self.nzeros,
                interpmethod=self.interpmethod,
                simpleoutput=True,
            ).transform(neaifg)

            spectrum_data[channelA] = amp
            spectrum_data[channelP] = phi
            spectrum_data["Wavenumber"] = f

        if self.simpleoutput:
            spectrum_data
        else:
            spectrum_parameters["ScanArea"] = [
                neaifg.parameters["ScanArea"][0],
                neaifg.parameters["ScanArea"][1],
                len(amp),
            ]
            spectrum = NeaSpectrum(spectrum_data, spectrum_parameters)
            return spectrum


class ProcessAllPoints(Transformation):
    def __init__(
        self,
        method="complex",
        apod=True,
        windowtype="blackmanharris",
        nzeros=4,
        interpmethod="spline",
    ):
        self.method = method
        self.apod = apod
        self.windowtype = windowtype
        self.nzeros = nzeros
        self.interpmethod = interpmethod

    def transform(self, neaifg):
        if (
            neaifg.parameters["PixelArea"][0] == 1
            and neaifg.parameters["PixelArea"][1] == 1
        ):
            spectra = ProcessMultiChannels(
                method=self.method,
                windowtype=self.windowtype,
                nzeros=self.nzeros,
                apod=self.apod,
                interpmethod=self.interpmethod,
                simpleoutput=False,
            ).transform(neaifg)
        else:
            pixel_area = [
                neaifg.parameters["PixelArea"][0],
                neaifg.parameters["PixelArea"][1],
                int(self.nzeros * neaifg.parameters["PixelArea"][2] / 2),
            ]

            ampFullData = np.zeros((pixel_area[0], pixel_area[1], pixel_area[2]))
            phiFullData = np.zeros(np.shape(ampFullData))
            fFullData = np.zeros(np.shape(ampFullData))

            pointifg_data = dict()
            pointifg_params = dict()
            pointifg_params["PixelArea"] = [1, 1, neaifg.parameters["PixelArea"][2]]
            pointifg_params["Scan"] = "Fourier Scan"
            pointifg_params["Averaging"] = neaifg.parameters["Averaging"]

            spectra_params = dict()
            spectra_params["Scan"] = "Fourier Scan"
            spectra_params["PixelArea"] = pixel_area
            spectra = NeaSpectrum({}, spectra_params, scantype=neaifg.scantype)

            allchannels = list(neaifg.data.keys())
            optical_channels = [
                name
                for name in allchannels
                if re.match("O(.?)A", name) or re.match("O(.?)P", name)
            ]
            orders = [int(n) for c in optical_channels for n in re.findall(r"\d", c)]
            orders = np.unique(np.asarray(orders))

            for order in orders:
                channelA = f"O{order}A"
                channelP = f"O{order}P"

                if channelA not in list(neaifg.data.keys()) or channelP not in list(
                    neaifg.data.keys()
                ):
                    print(
                        f"Skipped processing for order: {order}, since A or P is missing!"
                    )
                    continue
                else:
                    for i in range(pixel_area[0]):
                        for k in range(pixel_area[1]):
                            pointifg_data[channelA] = neaifg.data[channelA][i, k, :]
                            pointifg_data[channelP] = neaifg.data[channelP][i, k, :]
                            pointifg_data["M"] = neaifg.data["M"][i, k, :]
                            pointifg = NeaInterferogram(pointifg_data, pointifg_params)
                            (
                                ampFullData[i, k, :],
                                phiFullData[i, k, :],
                                fFullData[i, k, :],
                            ) = ProcessSingleChannel(
                                order,
                                method=self.method,
                                apod=self.apod,
                                windowtype=self.windowtype,
                                nzeros=self.nzeros,
                                interpmethod=self.interpmethod,
                                simpleoutput=True,
                            ).transform(
                                pointifg
                            )

                    spectra.data[channelA] = ampFullData
                    spectra.data[channelP] = phiFullData
                    spectra.data["Wavenumber"] = fFullData

        return spectra


# TOOLS ------------------------------------------------------------------------------------------------------------------
class Tools:
    def __init__(self):
        pass

    @staticmethod
    def reshape_ifg_data(data, params):
        if params["PixelArea"][1] != 1 or params["PixelArea"][0] != 1:
            for channel in list(data.keys()):
                data[channel] = np.reshape(
                    data[channel],
                    (
                        int(params["PixelArea"][0]),
                        int(params["PixelArea"][1]),
                        int(params["PixelArea"][2] * params["Averaging"]),
                    ),
                )
            return data
        else:
            return data

    @staticmethod
    def reshape_linescan_interferogram(data, parameters):
        return np.reshape(
            np.ravel(data), (parameters["PixelArea"][0], parameters["PixelArea"][2])
        )

    @staticmethod
    def asymmetric_window(npoints, centerindex=None, windowtype="blackmanharris"):
        if centerindex is None:
            centerindex = int(len(windowPart2) / 2)

        # Calculate the length of the two sides
        length1 = (centerindex) * 2
        length2 = (npoints - centerindex) * 2

        # Generate the two parts of the window

        windowfunc = getattr(signal.windows, windowtype)
        windowPart1 = windowfunc(length1)
        windowPart2 = windowfunc(length2)

        # Construct the asymetric window from the two sides
        asymWindow1 = windowPart1[0 : int(len(windowPart1) / 2)]
        if npoints % 2 == 0:
            asymWindow2 = windowPart2[int(len(windowPart2) / 2) : int(len(windowPart2))]
        else:
            asymWindow2 = windowPart2[int(len(windowPart2) / 2) : int(len(windowPart2))]

        return np.concatenate((asymWindow1, asymWindow2))

    @staticmethod
    def analyse_steps(maxis):
        stepsize = np.zeros((np.shape(maxis)[0], 1))
        stepspread = np.zeros((np.shape(maxis)[0], 1))
        for i in range(np.shape(maxis)[0]):
            stepsize[i] = np.mean(np.diff(maxis[i, :]))
            stepspread[i] = np.std(np.diff(maxis[i, :]))

        return stepsize, stepspread



"""
Module pySNOM
=============
Scanning Near-field Optical Microscopy (SNOM) analysis tools
"""
from .defaults import defaults
from .images import Image, Transformation
from .readers import Reader
from .spectra import NeaSpectrum
from .interferograms import NeaInterferogram

__version__ = "0.2.1"
__author__ = "Gergely Németh, Ferenc Borondics"
__credits__ = "Wigner Research Centre for Physics, Synchrotron SOLEIL"
__all__ = [
    "defaults",
    "Reader",
    "Image",
    "Transformation",
    "NeaSpectrum",
    "NeaInterferogram",
]



""" This file contains a dictionary of definitions to create connection
between the manufacturer nomeclature and the naming used in the package """


class Defaults:
    # Neaspec names are taken from the Scan field of the info txt file
    def __init__(self) -> None:
        self.image_mode_defs = {
            "AFM": "AFM",
            "2D (PsHet)": "PsHet",
            "Whitelight Imaging": "WLI",
            "Photo Thermal Expansion+": "PTE",
            "Tapping AFM-IR+": "TappingAFMIR",
            "Contact Mode 2D": "ContactAFM",
        }

        self.spectral_mode_defs = {
            "Fourier Scan": "nanoFTIR",
            "Pointspectroscopy PTE+": "PTE",
            "AFM-Raman/PL Scan (Tapping Mode)": "nanoRaman",
        }


defaults = Defaults()



import unittest
import os

import numpy as np

import pySNOM
from pySNOM import readers


class TestReaders(unittest.TestCase):
    def test_gwyread(self):
        f = "datasets/testPsHetData.gwy"
        file_reader = readers.GwyReader(
            fullfilepath=os.path.join(pySNOM.__path__[0], f), channelname="O3A raw"
        )
        gwy_image = file_reader.read()

        np.testing.assert_almost_equal(gwy_image.data[7][7], 15.213262557983398)
        np.testing.assert_almost_equal(gwy_image.data[8][8], 15.736936569213867)
        np.testing.assert_almost_equal(gwy_image.data[9][9], 13.609171867370605)

    def test_readgsf(self):
        f = "datasets/testPsHet O3A raw.gsf"
        file_reader = readers.GsfReader(os.path.join(pySNOM.__path__[0], f))
        gwy_image = file_reader.read()

        np.testing.assert_almost_equal(gwy_image.data[7][7], 15.213262557983398)
        np.testing.assert_almost_equal(gwy_image.data[8][8], 15.736936569213867)
        np.testing.assert_almost_equal(gwy_image.data[9][9], 13.609171867370605)

    def test_inforead(self):
        f = "datasets/testinfofile.txt"
        file_reader = readers.NeaInfoReader(os.path.join(pySNOM.__path__[0], f))
        data = file_reader.read()

        np.testing.assert_approx_equal(data["ScanArea"][0], 5.000)
        np.testing.assert_almost_equal(data["TargetWavelength"], 6.006)
        np.testing.assert_string_equal(data["DemodulationMode"], "PsHet")

    def test_general_reader_ifg(self):
        f = "datasets/testifg_singlepoint.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        np.testing.assert_almost_equal(data["O2A"][0], 9.580825)
        np.testing.assert_almost_equal(params["Regulator"][0], 3.767854)
        np.testing.assert_string_equal(params["Scan"], "Fourier Scan")

    def test_general_reader_spectrum(self):
        f = "datasets/testspectrum_singlepoint.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        np.testing.assert_almost_equal(data["O2A"][0], 0.1600194)
        np.testing.assert_string_equal(params["Scan"], "Fourier Scan")

    def test_general_reader_nanoraman(self):
        f = "datasets/testspectrum_nanoraman.txt"
        file_reader = readers.NeaSpectralReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        np.testing.assert_almost_equal(data["Data"][0], 11.0)
        np.testing.assert_string_equal(
            params["Scan"], "AFM-Raman/PL Scan (Tapping Mode)"
        )

    def test_legacy_nea_reader(self):
        f = "datasets/neafile_test_ifg.nea"
        file_reader = readers.NeaFileLegacyReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        np.testing.assert_almost_equal(data["O2A"][0], 0.8251932)
        np.testing.assert_array_equal(list(data.keys())[-1], "M")
        np.testing.assert_string_equal(params["Scan"], "Fourier Scan")


if __name__ == "__main__":
    unittest.main()



import numpy as np
import copy
import re
from enum import Enum
from pySNOM.defaults import Defaults

from skimage.transform import warp
from skimage.registration import optical_flow_tvl1, phase_cross_correlation
from scipy.ndimage import fourier_shift
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import spsolve
from scipy.ndimage import generic_filter

MeasurementModes = Enum(
    "MeasurementModes",
    ["None", "AFM", "PsHet", "WLI", "PTE", "TappingAFMIR", "ContactAFM"],
)
DataTypes = Enum("DataTypes", ["Amplitude", "Phase", "Topography"])
ChannelTypes = Enum("ChannelTypes", ["None", "Optical", "Mechanical"])


# Full measurement data containing all the measurement channels
class Measurement:
    def __init__(self, data, filename=None, info=None, mode="None"):
        self.filename = filename
        self.mode = mode
        self._data = data
        self.info = info

    @property
    def mode(self):
        """Property - measurement mode (Enum)"""
        return self._mode

    @mode.setter
    def mode(self, value: str):
        try:
            self._mode = MeasurementModes[value]
        except ValueError:
            self._mode = MeasurementModes["AFM"]
            raise ValueError(value + "is not a measurement mode!")

    @property
    def data(self):
        """Property - data (dict with GwyDataFields)"""
        return self._data

    @property
    def info(self):
        """Property - info (dictionary)"""
        return self._info

    @info.setter
    def info(self, info):
        self._info = info
        if not info == None:
            m = self._info["Scan"]
            self.mode = Defaults().image_mode_defs[m]

    # METHODS --------------------------------------------------------------------------------------
    def extract_channel(self, channelname: str):
        """Returns a single data channel as GwyDataField"""
        channel = self.data[channelname]
        return channel

    def image_from_channel(self, channelname: str):
        """Returns a single Image object with the requred channeldata"""
        channeldata = self.extract_channel(channelname)
        singleimage = GwyImage(
            channeldata,
            filename=self.filename,
            mode=self.mode,
            channel=channelname,
            info=self.info,
        )

        return singleimage


# Single image from a single data channel
class Image(Measurement):
    def __init__(
        self,
        data,
        filename=None,
        mode="AFM",
        channelname="Z raw",
        order=0,
        datatype=DataTypes["Topography"],
        info=None,
    ):
        super().__init__(data, filename, info=info, mode=mode)
        # Describing channel and datatype
        self.channel = channelname  # Full channel name
        self.order = int(order)  # Order, nth
        self.datatype = datatype  # Amplitude, Phase, Topography - Enum DataTypes

        self.data = data
        self.xoff = 0
        self.yoff = 0
        self.xreal = 1
        self.yreal = 1

    @property
    def data(self):
        """Property - data (numpy array)"""
        # Set the data
        return self._data

    @data.setter
    def data(self, new_data):
        """Setter for data (optional if you want data to be modifiable later)"""
        self._data = new_data
        self.xres, self.yres = np.shape(new_data)

    @property
    def channel(self):
        """Property - channel (string)"""
        return self._channel

    @channel.setter
    def channel(self, value):
        self._channel = value
        self.channeltype, self.order, self.datatype = type_from_channelname(value)


class GwyImage(Image):
    def __init__(
        self,
        data,
        filename=None,
        mode="AFM",
        channelname="Z raw",
        order=0,
        datatype=DataTypes["Topography"],
        info=None,
    ):
        super().__init__(
            data,
            filename=filename,
            mode=mode,
            channelname=channelname,
            order=order,
            datatype=datatype,
            info=info,
        )

        self.data = data
        self.xoff = data.xoff
        self.yoff = data.yoff
        self.xreal = data.xreal
        self.yreal = data.yreal

    @property
    def data(self):
        """Property - data (numpy array)"""
        # Set the data
        return self._data

    @data.setter
    def data(self, value):
        self._data = value.data
        self.xres, self.yres = np.shape(self._data)
        if self._data is None:
            raise ValueError(
                "The provided data object does not contain 'data' attribute"
            )


def type_from_channelname(channelname):
    channel_strings = ["M(.?)A", "M(.?)P", "O(.?)A", "O(.?)P", "Z C", "Z raw"]
    for pattern in channel_strings:
        if re.search(pattern, channelname) is not None:
            channel_name = re.search(pattern, channelname)[0]

    if channel_name[0] == "O":
        channeltype = ChannelTypes["Optical"]
    elif "M" in channel_name:
        channeltype = ChannelTypes["Mechanical"]
    else:
        channeltype = ChannelTypes["None"]

    if "Z" in channel_name:
        order = 0
    else:
        order = int(channel_name[1])

    if channel_name[2] == "A":
        datatype = DataTypes["Amplitude"]
    elif "Z" in channel_name:
        datatype = DataTypes["Topography"]
    else:
        datatype = DataTypes["Phase"]

    return channeltype, order, datatype


class Transformation:
    def transform(self, data):
        raise NotImplementedError()


class MaskedTransformation(Transformation):
    def calculate(self, data, mask=None):
        raise NotImplementedError()

    def correct(self, data, correction):
        """Applies the calculated corrections to the data"""
        if self.datatype == DataTypes.Amplitude:
            return data / correction
        else:
            return data - correction

    def transform(self, data, mask=None):
        """Calculates and applies the corrections to the data taking into account the mask if given"""
        correction = self.calculate(data, mask=mask)
        return self.correct(data, correction)


class LineLevel(MaskedTransformation):
    def __init__(self, method="median", datatype=DataTypes.Phase):
        self.method = method
        self.datatype = datatype

    def calculate(self, data, mask=None):
        if mask is not None:
            data = mask * data

        if self.method == "median":
            norm = np.nanmedian(data, axis=1, keepdims=True)
        elif self.method == "mean":
            norm = np.nanmean(data, axis=1, keepdims=True)
        elif self.method == "difference":
            if self.datatype == DataTypes.Amplitude:
                diff = data[1:] / data[:-1]
                diff = np.insert(diff, 0, 1, axis=0)
                norm = np.nancumprod(np.nanmedian(diff, axis=1))
            else:
                diff = data[1:] - data[:-1]
                diff = np.insert(diff, 0, 0, axis=0)
                norm = np.nancumsum(np.nanmedian(diff, axis=1))
            norm = np.tile(norm, (data.shape[1], 1)).T

        else:
            if self.datatype == DataTypes.Amplitude:
                norm = 1
            else:
                norm = 0

        return norm


class RotatePhase(Transformation):
    def __init__(self, degree=0.0):
        self.degree = degree

    def transform(self, data):
        # Construct complex dataset
        complexdata = np.exp(data * complex(1j))
        # Rotate and extract phase
        return np.angle(complexdata * np.exp(np.deg2rad(self.degree) * complex(1j)))


class SelfReference(Transformation):
    def __init__(self, referencedata=1, datatype=DataTypes.Phase):
        self.datatype = datatype
        self.referencedata = referencedata

    def transform(self, data):
        if self.datatype == DataTypes.Amplitude:
            return data / self.referencedata
        elif self.datatype == DataTypes.Phase:
            return data - self.referencedata
        else:
            raise RuntimeError(
                "Self-referencing makes only sense for amplitude or phase data"
            )


class SimpleNormalize(MaskedTransformation):
    def __init__(self, method="median", value=1.0, datatype=DataTypes.Phase):
        self.method = method
        self.value = value
        self.datatype = datatype

    def calculate(self, data, mask=None):
        """Calculates and returns the image corrections using mask (if given) without applying it to the data"""
        if mask is not None:
            data = mask * data

        match self.method:
            case "median":
                norm = np.nanmedian(data)
            case "mean":
                norm = np.nanmean(data)
            case "manual":
                norm = self.value
            case "min":
                norm = np.nanmin(data)

        return norm


class BackgroundPolyFit(Transformation):
    def __init__(self, xorder=1, yorder=1, datatype=DataTypes.Phase):
        self.xorder = int(xorder)
        self.yorder = int(yorder)
        self.datatype = datatype

    def calculate(self, data):
        """Calculates and returns the fitted polynomial background using mask (if given) without applying it to the data"""

        Z = copy.deepcopy(data)
        x = list(range(0, Z.shape[1]))
        y = list(range(0, Z.shape[0]))
        X, Y = np.meshgrid(x, y)
        x, y = X.ravel(), Y.ravel()
        b = Z.ravel()
        notnanidxs = np.argwhere(~np.isnan(b))
        b = np.ravel(b[notnanidxs])
        x = np.ravel(x[notnanidxs])
        y = np.ravel(y[notnanidxs])

        def get_basis(x, y, max_order_x=1, max_order_y=1):
            """Return the fit basis polynomials: 1, x, x^2, ..., xy, x^2y, ... etc."""
            basis = []
            for i in range(max_order_y + 1):
                # for j in range(max_order_x - i +1):
                for j in range(max_order_x + 1):
                    basis.append(x**j * y**i)
            return basis

        try:
            basis = get_basis(x, y, self.xorder, self.yorder)
            A = np.vstack(basis).T
            c, r, rank, s = np.linalg.lstsq(A, b, rcond=None)

            background = np.sum(
                c[:, None, None]
                * np.array(get_basis(X, Y, self.xorder, self.yorder)).reshape(
                    len(basis), *X.shape
                ),
                axis=0,
            )

        except ValueError:
            background = np.ones(np.shape(data))
            print("X and Y order must be integer!")

        return background

    def transform(self, data):
        """Calculates and applies the corrections to the data taking into account the mask if given"""
        background = self.calculate(data)

        if self.datatype == DataTypes["Amplitude"]:
            return data / background, background
        else:
            return data - background, background


class MaskedBackgroundPolyFit(BackgroundPolyFit, MaskedTransformation):
    def __init__(self, xorder=1, yorder=1, datatype=DataTypes.Phase):
        self.xorder = int(xorder)
        self.yorder = int(yorder)
        self.datatype = datatype

    def calculate(self, data, mask=None):
        """Calculates and returns the fitted polynomial background using mask (if given) without applying it to the data"""
        if mask is not None:
            data = mask * data

        return BackgroundPolyFit.calculate(self, data)

    def transform(self, data, mask=None):
        return MaskedTransformation.transform(self, data, mask=mask)

class LaplaceFillIn(Transformation):
    """
        Fill in missing (masked) regions of data using inward 
        interpolation via Laplace's equation. Handles edge and corner cases.
        Original NATLAB code: https://github.com/EvanCzako/image-spike-removal/blob/master/remove_spikes.m
        """
    
    def __init__(self, mask):  
        self.mask = mask
        
    def transform(self,data):
        """
        Parameters:
        - data (2D np.ndarray): Input data.
        - mask (2D np.ndarray): Boolean mask where True indicates missing values to fill.
        
        Returns:
        - filled (2D np.ndarray): Image with missing values filled.
        """

        M, N = data.shape
        num_pixels = M * N

        # Flattened indices
        u = np.flatnonzero(self.mask)      # masked (unknown) pixels
        w = np.flatnonzero(~self.mask)     # known pixels

        # Neighbor index offsets
        u_north = u - 1
        u_north = np.where(u % M != 0, u_north, 0)  # Wrap prevention for top row
        u_east = u + M
        u_east = np.where(u_east < num_pixels, u_east, 0)
        u_south = u + 1
        u_south = np.where((u + 1) % M != 0, u_south, 0)
        u_west = u - M
        u_west = np.where(u_west >= 0, u_west, 0)

        a = np.stack([u_north, u_east, u_south, u_west], axis=1)
        b = (a > 0).astype(float)
        sum_b = b.sum(axis=1, keepdims=True)
        c = -b / np.maximum(sum_b, 1e-12)

        # Sparse matrix entries
        row_inds = np.concatenate([u, u, u, u, u])
        col_inds = np.concatenate([u, u_north, u_east, u_south, u_west])
        data_vals = np.concatenate([
            np.ones(len(u)),
            c[:, 0], c[:, 1], c[:, 2], c[:, 3]
        ])

        # Remove invalid entries
        valid = (col_inds >= 0) & (col_inds < num_pixels)
        row_inds = row_inds[valid]
        col_inds = col_inds[valid]
        data_vals = data_vals[valid]

        # Include identity rows for known pixels
        row_inds = np.concatenate([row_inds, w])
        col_inds = np.concatenate([col_inds, w])
        data_vals = np.concatenate([data_vals, np.ones(len(w))])

        # Build sparse matrix
        A = coo_matrix((data_vals, (row_inds, col_inds)), shape=(num_pixels, num_pixels)).tocsr()

        # Build RHS vector
        b_vec = data.flatten()
        b_vec[self.mask.flatten()] = 0

        # Solve linear system
        x = spsolve(A, b_vec)
        filled = x.reshape(data.shape)

        return filled
    
class ValueFillIn(Transformation):
    def __init__(self, mask, value):
        self.value = value
        self.mask = mask

    def transform(self, data):
        data[np.isnan(self.mask)] = self.value
        
        return data

class RemoveSpikes(Transformation):
    def __init__(self, threshold=0.8, absolute_threshold=False, method='laplace', higher=False, value = 1.0):
        self.threshold = threshold
        self.method = method
        self.higher = higher
        self.value = value
        self.absolute_threshold = absolute_threshold
        
    def transform(self, data):
        if not self.absolute_threshold:
            norm_data = data/np.nanmedian(data)
        else:
            norm_data = data

        if self.higher:
            spike_mask = mask_from_datacondition(norm_data > self.threshold)
        else:
            spike_mask = mask_from_datacondition(norm_data < self.threshold)
    
        # Use previously defined fill_region to fill spikes
        match self.method:
            case "laplace":
                data = LaplaceFillIn(np.isnan(spike_mask)).transform(data)
            case "median":
                data = ValueFillIn(spike_mask,np.nanmedian(data)).transform(data)
            case "manual":
                data = ValueFillIn(spike_mask,self.value).transform(data)
            case _:
                data = spike_mask*data

        return data


class ScarRemoval(Transformation):
    def __init__(self, threshold=0.5, flip=False, datatype=DataTypes.Phase):
        self.threshold = threshold
        self.flip = flip
        self.datatype = datatype

    def transform(self, data):
        d = copy.deepcopy(data)
        in_data = copy.deepcopy(data)
        if self.flip:
            d = d.T
            in_data = in_data.T

        for i in range(1, in_data.shape[0] - 1):
            b = in_data[i - 1, :]
            c = in_data[i, :]
            a = in_data[i + 1, :]
            scarmask = np.abs(b - a) < self.threshold * (np.abs(c - a))
            d[i, scarmask] = (b[scarmask] + a[scarmask]) / 2

        if self.flip:
            d = d.T
        return d


# TODO: Helper functions to create masks or turn other types of masks into 1/Nan mask
def mask_from_booleans(bool_mask, bad_values=False):
    """Turn a boolean array to an array conatining nans and ones"""
    mshape = np.shape(bool_mask)
    return np.where(bool_mask == bad_values, np.nan * np.ones(mshape), np.ones(mshape))


def mask_from_datacondition(condition):
    mshape = np.shape(condition)
    return np.where(condition, np.nan * np.ones(mshape), np.ones(mshape))


class CalculateOpticalFlow(Transformation):
    """Calculates the pixel coordiate drifts between reference and template image"""

    def __init__(self, image_ref):
        self.image_ref = image_ref

    def transform(self, image):
        v, u = optical_flow_tvl1(
            self.image_ref / np.nanmax(self.image_ref), image / np.nanmax(image)
        )
        return v, u


class WrapImage(Transformation):
    """Applies the pixel-by-pixel drift correction calculated by OpticalFlow"""

    def __init__(self, v, u):
        self.v = v
        self.u = u

    def transform(self, image):
        nr, nc = image.shape
        row_coords, col_coords = np.meshgrid(
            np.arange(nr), np.arange(nc), indexing="ij"
        )
        return warp(
            image, np.array([row_coords + self.v, col_coords + self.u]), mode="edge"
        )


class CalculateXCorrDrift(Transformation):
    """Calculates the drift between reference and template image"""

    def __init__(self, image_ref):
        self.image_ref = image_ref

    def transform(self, image):
        shift, _, _ = phase_cross_correlation(self.image_ref, image)
        return shift


class CorrectImageDrift(Transformation):
    """Rearranges image pixels to correct image shift calculated by cross-correlation"""

    def __init__(self, shift):
        self.shift = shift

    def transform(self, image):
        offset_phase = fourier_shift(np.fft.fftn(image), self.shift)
        offset_phase = np.fft.ifftn(offset_phase)
        return offset_phase.real


class AlignImageStack(Transformation):
    """Calculates the drift between the given images and organize the comman areas into an aligned stack"""

    def __init__(self):
        pass

    def calculate(self, images):
        shifts = []
        crossrect = [0, 0, np.shape(images[0])[0], np.shape(images[0])[1]]
        if len(images) > 1:
            xcorr = CalculateXCorrDrift(images[0])
            for i in range(len(images)):
                if i > 0:
                    shifts.append(xcorr.transform(images[i]))
                    crossrect = shifted_cross_section(
                        rect1=crossrect,
                        rect2=[
                            -shifts[-1][0],
                            shifts[-1][1],
                            np.shape(images[i])[0],
                            np.shape(images[i])[1],
                        ],
                    )
            return shifts, crossrect
        else:
            return None

    def transform(self, images, shifts, crossrect):
        aligned_stack = []
        for i in range(len(images)):
            if i > 0:
                shifter = CorrectImageDrift(shifts[i - 1])
                aligned_stack.append(shifter.transform(images[i]))
                aligned_stack[i] = cut_image(aligned_stack[i], crossrect)
            else:
                aligned_stack.append(cut_image(images[i], crossrect))
        return aligned_stack


def sort_image_stack(images, wns):
    """Sort the image stack based on the wavenumber list"""

    idxs = np.argsort(np.asarray(wns))
    images = [images[i] for i in idxs]
    wns = [wns[i] for i in idxs]

    return images, wns


def create_nparray_stack(measlist):
    """Creates a numpy array stack from a list of measurements, organized as [ rows, columns, wavelengths ] (compatible with quasar io utils)"""

    stack = np.zeros(
        (np.shape(measlist[0])[0], np.shape(measlist[0])[1], len(measlist))
    )

    for i, meas in enumerate(measlist):
        stack[:, :, i] = meas

    return stack


def dict_from_imagestack(X, channelname, wn=None, is_interferogram=True):
    """Converts the image stack into a pySNOM spectra or interferograms compatible dictionary"""
    final_dict = {}
    params = {}

    X = np.asarray(X)

    params["PixelArea"] = [X.shape[1], X.shape[2], X.shape[0]]
    params["Averaging"] = 1
    params["Scan"] = "Fourier Scan"

    final_dict[channelname] = flatten_stack(X)

    y_loc = np.repeat(np.arange(X.shape[1]), X.shape[2])
    x_loc = np.tile(np.arange(X.shape[2]), X.shape[1])

    final_dict["Row"] = np.repeat(y_loc, X.shape[0])
    final_dict["Column"] = np.repeat(x_loc, X.shape[0])

    if is_interferogram:
        depth_channel_name = "M"
    else:
        depth_channel_name = "Wavenumber"

    if wn is not None:
        final_dict[depth_channel_name] = np.tile(wn, X.shape[1] * X.shape[2])
    else:
        final_dict[depth_channel_name] = np.tile(
            np.arange(X.shape[0]), X.shape[1] * X.shape[2]
        )

    return final_dict, params


def flatten_stack(imagestack):
    """Flatten out values in an image stack to be aneble to add it to spectral dictionaries"""
    imagestack = np.asarray(imagestack)
    flattened_values = imagestack.reshape(
        (imagestack.shape[0], imagestack.shape[1] * imagestack.shape[2])
    )
    return np.ravel(flattened_values, order="F")


def shifted_cross_section(rect1: list, rect2: list):
    """Calculates the cross-section of two rectangle shifted to each other"""
    x1 = rect1[1]
    x2 = rect2[1]
    y1 = rect1[0]
    y2 = rect2[0]
    W1 = rect1[3]
    W2 = rect2[3]
    H1 = rect1[2]
    H2 = rect2[2]

    if y2 > y1:
        Hn = H1 - (y2 - y1)
        yn = y2
    elif (y2 < y1) and (y1 + H1 > y2 + H2):  # Negative shift and higher than H2
        Hn = H2 + (y2 - y1)
        yn = y1
    else:
        Hn = H1
        yn = y1

    if x2 > x1:  # Positive shift
        Wn = W1 - (x2 - x1)
        xn = x2
    elif (x2 < x1) and (x1 + W1 > x2 + W2):  # Negative shift and higher than W2
        Wn = W2 + (x2 - x1)
        xn = x1
    else:
        Wn = W1
        xn = x1

    return int(yn), int(xn), int(Hn), int(Wn)


def cut_image(image, rect):
    """Cuts the part of the image array defined by rectangle"""
    return image[-(rect[2]) : -(rect[0] + 1), rect[1] : rect[1] + rect[3]]
